/* hash_serialize.c - stringify and destringify a GLib GHashTable
 * Copyright 2002 by Gaal Yahas <gaal@forum2.org>
 */

#include <gtk/gtk.h>
#include <glib.h>
#include <stdlib.h>
#include <errno.h>
#include "hash_serialize.h"

//static void _hash_element_freeze(gpointer key, gpointer val, SerializationData *sd);

static const char SERIALIZE_DELIMITER = '%';

static void
_hash_element_freeze(gpointer key, gpointer val, SerializationData *sd) {
	GString *cleankey = sd->key_serialize(key);
	GString *cleanval = sd->val_serialize(val);

	g_string_append_printf(sd->accum, "%s%c%s%c",
		cleankey->str, SERIALIZE_DELIMITER, cleanval->str, SERIALIZE_DELIMITER);
}

/* stringify a GHashTable. Knowing the hash's type and connecting
 * the appropriate serializing functions is entirely the user's
 * responsibility. */
GString*
lj_hash_freeze(GHashTable* hash,
		SerializeFunc key_serialize,
		SerializeFunc val_serialize) {

	SerializationData sd = { key_serialize, val_serialize, NULL };

	sd.accum = g_string_new("");

	g_hash_table_foreach(hash, (GHFunc)_hash_element_freeze, &sd);

	return sd.accum;
}

GString*
gint_serialize(gint* data) {
	GString *str = g_string_new("");
	g_string_printf(str, "%d", *data);
	return str;
}

/* XXX: This serializer is not Unicode ready */
GString*
gchar_serialize(gchar* data) {
	GString *str = g_string_sized_new(strlen(data));
	gchar *p;

	for (p = data; *p; p++) {
		g_string_append_c(str, *p);
		if (*p == SERIALIZE_DELIMITER)
			g_string_append_c(str, SERIALIZE_DELIMITER);
	}
	return str;
}

/* Build a GHashTable from one frozen into a GString with lj_hash_freeze().
 *
 * The user must create an empty GHashTable with appropriate hash and
 * comparison functions before calling this function. Also, it is up to
 * the user to connect the right deserialization functions for both the
 * hash key and value. */
void
lj_hash_thaw(GHashTable* hash,
		SerializeFunc key_deserialize,
		SerializeFunc val_deserialize,
		GString *frozen_hash) {

	GString *frozen_key = g_string_new("");
	GString *frozen_val = g_string_new("");

	GString *current = frozen_key;
	gpointer key, val;
	gint len = frozen_hash->len;
	gint i;

	/* some integrity checks */
	if (frozen_hash->len == 0) return;
	/* if (frozen_hash->str[0] == SERIALIZE_DELIMITER)
		g_error("lj_hash_thaw() called but 0==SD");
	if (frozen_hash->str[len-1] != SERIALIZE_DELIMITER)
		g_error("lj_hash_thaw() called but -1!=SD"); */
	if ((frozen_hash->str[0] == SERIALIZE_DELIMITER) ||
		(frozen_hash->str[len-1] != SERIALIZE_DELIMITER)) {

		g_error("lj_hash_thaw() called with non-frozen data (%p)", frozen_hash);
		//g_warning("lj_hash_thaw() called with non-frozen data", frozen_hash);
		//return;
	}

	for (i = 0; i < len; i++) {
		if (frozen_hash->str[i] == SERIALIZE_DELIMITER) {
			if ((i == len - 1) || (frozen_hash->str[i+1] != SERIALIZE_DELIMITER)) {
				if (current == frozen_val) {
					key = key_deserialize(frozen_key);
					val = val_deserialize(frozen_val);

					g_hash_table_insert(hash, key, val);

					//g_free(key);
					//g_free(val);

					frozen_val = g_string_truncate(frozen_val, 0);
					frozen_key = g_string_truncate(frozen_key, 0);

					current = frozen_key;
				} else {
					current = frozen_val;
				}
			} else {
				g_string_append_c(current, SERIALIZE_DELIMITER);
				i++;
			}
		} else {
			g_string_append_c(current, frozen_hash->str[i]);
		}
	}

	g_string_free(frozen_key, TRUE);
	g_string_free(frozen_val, TRUE);
}


gint*
gint_deserialize(GString* data) {
	gint* i   = g_new(gint, 1);
	glong tmp = strtol(data->str, (char **)NULL, 10);

	if (errno)
		g_error("conversion error in gint_deserialize");

	*i = (gint)tmp;

	return i;
}

gchar*
gchar_deserialize(GString* data) {
	return g_strdup(data->str);
}
